@@ -276,6 +276,40 @@ _make_option(
 
 =pod
 
+=head3 B<lists_can_interrupt_paragraph> I<(enum, default: strict)>
+
+Specify whether and how a list can interrupt a paragraph.
+
+=over 4
+
+=item B<never>
+
+A list can never interrupt a paragraph.
+
+=item B<within_list>
+
+A list can interrupt a paragraph only when we are already inside another list.
+
+=item B<strict> I<(default)>
+
+A list can interrupt a paragraph but only with some non ambiguous list markers.
+
+=item B<always>
+
+A list can always interrupt a paragraph.
+
+=back
+
+=cut
+
+_make_option(
+  lists_can_interrupt_paragraph => 'strict',
+  _enum(qw(never within_list strict always)), (
+    markdown => 'within_list',
+  ));
+
+=pod
+
 =head3 B<allow_task_list_markers> I<(enum, default: list)>
 
 Specify whether task list markers (rendered as check boxes) are recognised in
@@ -388,6 +422,18 @@ _make_option(code_blocks_info => 'language', _enum(qw(ignored language)));
 
 =pod
 
+=head3 B<code_blocks_convert_tabs_to_spaces> I<(boolean, default: false)>
+
+By default, tabs are preserved inside code blocks. With this option, all tabs (at
+the beginning of the lines or inside) are turned into spaces, aligned with the
+tab stops (currently always a multiple of 4).
+
+=cut
+
+_make_option(code_blocks_convert_tabs_to_spaces => 0, _boolean, (markdown => 1));
+
+=pod
+
 =head3 B<table_blocks_have_cells_for_missing_data> I<(boolean, default: false)>
 
 Whether a table will have a cell in HTML for a missing cell in the markdown
@@ -599,18 +645,6 @@ _make_option(force_final_new_line => 0, _boolean, (markdown => 1));
 
 =pod
 
-=head3 B<preserve_tabs> I<(boolean, default: true)>
-
-When removing prefix spaces in front of some constructs (typically indented code
-blocks), pmarkdown will try to preserve tabs when they are used instead of
-space. If this option is set to false, prefix tabs will be turned into spaces.
-
-=cut
-
-_make_option(preserve_tabs => 1, _boolean, (markdown => 0));
-
-=pod
-
 =head3 B<preserve_white_lines> I<(boolean, default: true)>
 
 By default, pmarkdown will try to preserve lines that contains only whitespace
